include("../../AoC2021.jl")

##

function parsefile(file)

    points = []

    for line in eachline(file)
        if line == ""
            break
        end

        push!(points, (split(line, ",") .|>
                       Base.Fix1(parse, Int)) |>
                      x -> (x[1], x[2]).+1)
    end

    paper = zeros(Bool, maximum(map(((x,y),)->y, points)),maximum(map(((x,y),)->x, points)))

    for (x,y) ∈ points
        paper[y,x] = true
    end

    folds = map(eachline(file)) do line
        matches = match(r"fold along (?<axis>[xy])=(?<coord>\d+)", line)

        (matches["axis"] == "x" ? Val(x) : Val(y), parse(Int, matches["coord"])+1)
    end

    (paper, folds)
end

@enum Axes x y

function foldpaper(paper, ::Val{x}, coord)
    folded = paper[:, end:-1:(coord+1)]

    foldedpaper = paper[:, 1:(coord-1)]
    foldedpaper[:, (end-size(folded, 2)+1):end] .|= folded

    foldedpaper
end

function foldpaper(paper, ::Val{y}, coord)
    folded = paper[end:-1:(coord+1), :]


    foldedpaper = paper[1:(coord-1), :]
    foldedpaper[(end-size(folded, 1)+1):end, :] .|= folded

    foldedpaper
end

##

function task0(file)
    (paper, folds) = parsefile(file)

    sum(foldpaper(paper, folds[1]...))
end

##

verify(13, 1, 0, 17)
verify(13, 1, 1, 693)

##

function task1(file)
    (paper, folds) = parsefile(file)

    folded = foldl((paper, foldins) -> foldpaper(paper, foldins...), folds, init=paper)

    eachrow(map(x-> x ? "█" : " ", folded)) .|>
    (x->join(x,"")) |>
    (x->join(x,'\n')) |> println

    folded
end

##

verify(13, 2, 0, 
[
    1  1  1  1  1
    1  0  0  0  1
    1  0  0  0  1
    1  0  0  0  1
    1  1  1  1  1
    0  0  0  0  0
    0  0  0  0  0
])
verify(13, 2, 1, 
[
    1  0  0  1  0  0  1  1  0  0  1  0  0  0  0  1  1  1  1  0  1  1  1  0  0  0  1  1  0  0  1  1  1  1  0  1  0  0  1  0
    1  0  0  1  0  1  0  0  1  0  1  0  0  0  0  0  0  0  1  0  1  0  0  1  0  1  0  0  1  0  0  0  0  1  0  1  0  0  1  0
    1  0  0  1  0  1  0  0  0  0  1  0  0  0  0  0  0  1  0  0  1  0  0  1  0  1  0  0  1  0  0  0  1  0  0  1  0  0  1  0
    1  0  0  1  0  1  0  0  0  0  1  0  0  0  0  0  1  0  0  0  1  1  1  0  0  1  1  1  1  0  0  1  0  0  0  1  0  0  1  0
    1  0  0  1  0  1  0  0  1  0  1  0  0  0  0  1  0  0  0  0  1  0  1  0  0  1  0  0  1  0  1  0  0  0  0  1  0  0  1  0
    0  1  1  0  0  0  1  1  0  0  1  1  1  1  0  1  1  1  1  0  1  0  0  1  0  1  0  0  1  0  1  1  1  1  0  0  1  1  0  0
])
